/*
    Tryton Android
    Copyright (C) 2012 SARL SCOP Scil (contact@scil.coop)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package org.tryton.client.models;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.TreeMap;
import java.util.Map;
import java.util.Set;

/** This is just an aggregation of views per type for a model name. 
 * It holds loaded views and allows to explicitely define ids of views to
 * load in given types.
 * If either type or even ModelViewTypes is not set for a field, it means
 * that default views should be used. Also if a type is set to use the id 0
 * it means that the default view should be used. Once set the actual id
 * is updated. */
public class ModelViewTypes implements Serializable {

    /** Autogenerated serial UID */
    static final long serialVersionUID = 4878345939938015119L;

    private String modelName;
    private Map<String, Integer> viewIds;
    private Map<String, ModelView> views;

    public ModelViewTypes(String modelName) {
        this.modelName = modelName;
        this.viewIds = new TreeMap<String, Integer>();
        this.views = new TreeMap<String, ModelView>();
    }
    
    /** Add a subview id, without view.
     * Use 0 to explicitely use default view. */
    public void putViewId(String type, int id) {
        this.viewIds.put(type, id);
    }
    /** Add a subview (also add its id) */
    public void putView(String type, ModelView view) {
        this.views.put(type, view);
        this.viewIds.put(type, view.getId());
    }

    public ModelView getView(String type) {
        return this.views.get(type);
    }

    /** Get the id of a view for a type. 0 means either use default one,
     * or that the type has not been explicitely set (thus use default). */
    public int getViewId(String type) {
        Integer id = this.viewIds.get(type);
        if (id == null) {
            return 0;
        } else {
            return id.intValue();
        }
    }

    public Set<String> getTypes() {
        return this.viewIds.keySet();
    }

    /** Check if a view is defined for the given type (either loaded or not) */
    public boolean hasView(String type) {
        return this.viewIds.containsKey(type);
    }

    public String getModelName() {
        return this.modelName;
    }

    /** Get the list off all fields, used in any of the types. Expects each
     * view to be loaded and built. */
    public List<String> getAllFieldNames() {
        List<String> fields = new ArrayList<String>();
        for (String type : this.getTypes()) {
            ModelView v = this.views.get(type);
            for (Model field : v.getStructure()) {
                String className = field.getClassName();
                String fieldName = field.getString("name");
                if (className.equals("label")) {
                    // ignore
                } else if (className.equals("graph.axis.x")
                           || className.equals("graph.axis.y")) {
                    // Graph axis, get subfields
                    @SuppressWarnings("unchecked")
                    List<Model> axisFields = (List<Model>) field.get("axis");
                    for (Model axisModel : axisFields) {
                        fieldName = axisModel.getString("name");
                        if (!fields.contains(fieldName)) {
                            fields.add(fieldName);
                        }
                    }
                } else  if (!fields.contains(fieldName)) {
                    // Standard field
                    fields.add(fieldName);
                }
            }
        }
        return fields;
    }

    public ModelViewTypes copy() {
        ModelViewTypes copy = new ModelViewTypes(this.modelName);
        for (String type : this.viewIds.keySet()) {
            copy.viewIds.put(type, this.viewIds.get(type));
            copy.views.put(type, this.views.get(type));
        }
        return copy;
    }
}